from threading import RLock
from typing import List

from bobocep.decider.buffers.shared_versioned_match_buffer import \
    SharedVersionedMatchBuffer
from bobocep.decider.handlers.abstract_handler import AbstractHandler
from bobocep.decider.handlers.nfa_handler_subscriber import \
    INFAHandlerSubscriber
from bobocep.decider.runs.bobo_run import BoboRun
from bobocep.decider.runs.run_subscriber import IRunSubscriber
from bobocep.receiver.clocks.epoch_ns_clock import EpochNSClock
from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.nfas.bobo_nfa import BoboNFA


class BoboNFAHandler(AbstractHandler, IRunSubscriber):
    """A :code:`bobocep` automaton handler.

    :param nfa: The automaton with which the handler is associated.
    :type nfa: BoboNFA

    :param buffer: The buffer in which the handler will store event data.
    :type buffer: SharedVersionedMatchBuffer

    :param max_recent: The maximum number of recent CompositeEvent instances
                        generated by this handler to pass to runs.
                        Minimum of 1, defaults to 1.
    :type max_recent: int, optional
    """

    NFA_NAME = "nfa_name"
    BUFFER = "buffer"
    RUNS = "runs"

    def __init__(self,
                 nfa: BoboNFA,
                 buffer: SharedVersionedMatchBuffer,
                 max_recent: int = 1) -> None:
        super().__init__()

        self.nfa = nfa
        self.buffer = buffer

        self._runs = {}
        self._recent = []
        self._max_recent = max(1, max_recent)
        self._subs = []
        self._lock = RLock()

    def process(self, event: BoboEvent) -> None:
        with self._lock:
            recent = self._recent[:]
            self._check_event_against_runs(event, recent)

            if self.nfa.start_state.process(event, {}, recent):
                # only notify if start state isn't also final state
                self.on_run_clone(
                    state_name=self.nfa.start_state.name,
                    event=event,
                    parent_run_id=None,
                    force_parent=False,
                    notify=True)

    def _check_event_against_runs(self,
                                  event: BoboEvent,
                                  recent: List[CompositeEvent]) -> None:
        for run in tuple(self._runs.values()):
            run.process(event, recent)

    def add_run(self, run: BoboRun) -> None:
        """
        Adds a run to the handler.

        :param run: The run.
        :type run: BoboRun
        """

        with self._lock:
            if run.id in self._runs:
                raise RuntimeError(
                    "Run ID {} already exists in handler for NFA {}.".format(
                        run.id, self.nfa.name))

            self._runs[run.id] = run
            run.subscribe(self)

    def remove_run(self,
                   run_id: str,
                   halt: bool = False,
                   notify: bool = False) -> None:
        """
        Removes a run from the handler.

        :param run_id: The ID of the run to remove.
        :type run_id: str

        :param halt: Halts the run first, defaults to False.
        :type halt: bool, optional

        :param notify: Notifies subscribers that the run has halted
                       *only if* halt is True. Defaults to False.
        :type notify: bool, optional
        """

        with self._lock:
            run = self._runs.get(run_id)

            if run is not None:
                if halt:
                    run.set_halt(notify=notify)

                # remove current run version from the pointers of all
                # match events in buffer
                self.buffer.remove_version(
                    nfa_name=self.nfa.name,
                    version=run.version.get_version_as_str())

                # remove run from handler
                self._runs.pop(run.id, None)

    def clear_runs(self,
                   halt: bool = True,
                   notify: bool = True) -> None:
        """
        Removes all runs from the handler.

        :param halt: Halts all runs first, defaults to False.
        :type halt: bool, optional

        :param notify: Notifies subscribers that runs have halted
                       *only if* halt is True. Defaults to False.
        :type notify: bool, optional
        """

        with self._lock:
            for run in tuple(self._runs.values()):
                self.remove_run(run_id=run.id,
                                halt=halt,
                                notify=notify)

    def on_run_transition(self,
                          run_id: str,
                          state_name_from: str,
                          state_name_to: str,
                          event: BoboEvent,
                          notify: bool = True) -> None:
        with self._lock:
            if notify:
                self._notify_transition(
                    run_id=run_id,
                    state_name_from=state_name_from,
                    state_name_to=state_name_to,
                    event=event)

    def on_run_clone(self,
                     state_name: str,
                     event: BoboEvent,
                     parent_run_id: str = None,
                     force_parent: bool = False,
                     notify: bool = True) -> None:
        """
        :raises RuntimeError: No parent run found when either one is specified
                              or force_parent is True.
        :raises RuntimeError: State name not found in automaton.
        """

        with self._lock:
            if parent_run_id is None:
                parent_run = None
            else:
                parent_run = self._runs.get(parent_run_id)

            if parent_run is None and \
                    (parent_run_id is not None or force_parent):
                raise RuntimeError(
                    "Failed to clone run. No parent run found.")

            next_state = self.nfa.states.get(state_name)

            if next_state is None:
                raise RuntimeError(
                    "Failed to clone run. "
                    "State {} not found in NFA {}.".format(
                        state_name, self.nfa.name))

            clone_run = BoboRun(
                buffer=self.buffer,
                nfa=self.nfa,
                event=event,
                start_state=next_state,
                parent_run=parent_run)

            self.add_run(clone_run)

            if clone_run.is_final():
                self.on_run_final(
                    run_id=clone_run.id,
                    history=BoboHistory({
                        clone_run.current_state.label: [clone_run.event]
                    }),
                    halt=True,
                    notify=notify
                )

            elif notify:
                self._notify_clone(
                    run_id=clone_run.id,
                    state_name=clone_run.current_state.name,
                    event=clone_run.event)

    def on_run_final(self,
                     run_id: str,
                     history: BoboHistory,
                     halt: bool = False,
                     notify: bool = True) -> None:
        """
        :raises RuntimeError: Run ID not found.
        """

        with self._lock:
            run = self._runs.get(run_id)

            if run is None:
                raise RuntimeError(
                    "Run {} not found in handler for NFA {}.".format(
                        run_id, self.nfa.name))

            if halt:
                run.set_halt(notify=False)

            event = CompositeEvent(
                timestamp=EpochNSClock.generate_timestamp(),
                name=self.nfa.name,
                history=history)

            if notify:
                self._notify_final(run_id, event)

            self._add_recent(event)
            self.remove_run(run_id)

    def on_run_halt(self,
                    run_id: str,
                    notify: bool = True) -> None:
        with self._lock:
            if notify:
                self._notify_halt(run_id)
            self.remove_run(run_id)

    def force_run_transition(self,
                             run_id: str,
                             state_name_from: str,
                             state_name_to: str,
                             event: BoboEvent) -> None:
        """
        Forces a state transition on a run without notification.

        :param run_id: The run ID.
        :type run_id: str

        :param state_name_from: The original state before the transition.
        :type state_name_from: str

        :param state_name_to: The new state after the transition.
        :type state_name_to: str

        :param event: The event that caused the state transition.
        :type event: BoboEvent

        :raises RuntimeError: Run ID not found.
        :raises RuntimeError: From state not found.
        :raises RuntimeError: To state not found.
        :raises RuntimeError: To state is not a transition for From state.
        :raises RuntimeError: From state is not the current run state.
        """

        with self._lock:
            run = self._runs.get(run_id)

            if run is None:
                raise RuntimeError(
                    "Run ID {} not found in handler for NFA {}.".format(
                        run_id, self.nfa.name))

            state_from = self.nfa.states.get(state_name_from)

            if state_from is None:
                raise RuntimeError(
                    "State {} not found in NFA {}.".format(
                        state_name_from, self.nfa.name))

            state_to = self.nfa.states.get(state_name_to)

            if state_to is None:
                raise RuntimeError(
                    "State {} not found in NFA {}.".format(
                        state_name_to, self.nfa.name))

            trans_from = self.nfa.transitions[state_name_from]

            if state_name_to not in trans_from.state_names:
                raise RuntimeError(
                    "State {} not a transition for state {}.".format(
                        state_name_to, state_name_from))

            if run.current_state.name != state_name_from:
                raise RuntimeError(
                    "Current state for run {} is {} and not {}.".format(
                        run_id, run.current_state.name, state_name_from))

            run._proceed(event=event,
                         original_state=state_from,
                         trans_state=state_to,
                         increment=state_name_from == state_name_to,
                         notify=False)

    def force_run_clone(self,
                        state_name: str,
                        event: BoboEvent,
                        parent_run_id: str) -> None:
        """
        Forces a run clone without notification.

        :param state_name: The state of the run after cloning.
        :type state_name: str

        :param event: The event that caused the clone.
        :type event: BoboEvent

        :param parent_run_id: The parent run ID.
        :type parent_run_id: str
        """

        with self._lock:
            self.on_run_clone(
                state_name=state_name,
                event=event,
                parent_run_id=parent_run_id,
                force_parent=False,
                notify=False)

            run = self._runs.get(parent_run_id)

            if run is not None:
                run._last_proceed_had_clone = True

    def force_run_halt(self, run_id: str) -> None:
        """
        Forces a halt without notification.

        :param run_id: The run ID.
        :type run_id: str

        :raises RuntimeError: Run ID not found.
        :raises RuntimeError: Run has already halted.
        """

        with self._lock:
            run = self._runs.get(run_id)

            if run is None:
                raise RuntimeError(
                    "Run {} not found in handler for NFA {}.".format(
                        run_id, self.nfa.name))

            if run.is_halted():
                raise RuntimeError(
                    "Run {} has already halted for NFA {}.".format(
                        run_id, self.nfa.name))

            run.set_halt(notify=False)
            self.on_run_halt(run_id, notify=False)

    def force_run_final(self, run_id: str, history: BoboHistory):
        """
        Forces a run to reach its final state without notification.

        :param run_id: The run ID.
        :type run_id: str

        :param history: The history of the accepted run.
        :type history: BoboHistory
        """

        with self._lock:
            self.on_run_final(run_id=run_id,
                              history=history,
                              halt=True,
                              notify=False)

    def subscribe(self, subscriber: INFAHandlerSubscriber) -> None:
        """
        :param subscriber: Subscribes to state transitions within handler.
        :type subscriber: INFAHandlerSubscriber
        """

        with self._lock:
            if subscriber not in self._subs:
                self._subs.append(subscriber)

    def unsubscribe(self, unsubscriber: INFAHandlerSubscriber) -> None:
        """
        :param unsubscriber: Unsubscribes from state transitions within
                             handler.
        :type unsubscriber: INFAHandlerSubscriber
        """

        with self._lock:
            if unsubscriber in self._subs:
                self._subs.remove(unsubscriber)

    def _add_recent(self, event: CompositeEvent) -> None:
        # add new event to the start of the list
        self._recent.insert(0, event)

        # ensure first event has the largest timestamp i.e. is most recent
        self._recent.sort(key=lambda e: e.timestamp, reverse=True)

        # delete oldest event if list length exceeds the maximum
        if len(self._recent) > self._max_recent:
            self._recent.pop()

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        with self._lock:
            return {
                self.NFA_NAME: self.nfa.name,
                self.BUFFER: self.buffer.to_dict(),
                self.RUNS: [run.to_dict() for run in self._runs.values()
                            if not run.is_halted()]
            }

    def _notify_transition(self,
                           run_id: str,
                           state_name_from: str,
                           state_name_to: str,
                           event: BoboEvent) -> None:
        for subscriber in self._subs:
            subscriber.on_handler_transition(
                nfa_name=self.nfa.name,
                run_id=run_id,
                state_name_from=state_name_from,
                state_name_to=state_name_to,
                event=event)

    def _notify_clone(self,
                      run_id: str,
                      state_name: str,
                      event: BoboEvent) -> None:
        for subscriber in self._subs:
            subscriber.on_handler_clone(
                nfa_name=self.nfa.name,
                run_id=run_id,
                state_name=state_name,
                event=event)

    def _notify_final(self, run_id: str, event: CompositeEvent) -> None:
        for subscriber in self._subs:
            subscriber.on_handler_final(
                nfa_name=self.nfa.name,
                run_id=run_id,
                event=event)

    def _notify_halt(self, run_id: str) -> None:
        for subscriber in self._subs:
            subscriber.on_handler_halt(
                nfa_name=self.nfa.name,
                run_id=run_id)



from bobocep.receiver.clocks.epoch_ns_clock import EpochNSClock
from bobocep.receiver.formatters.abstract_formatter import AbstractFormatter
from bobocep.rules.events.primitive_event import PrimitiveEvent


class PrimitiveEventFormatter(AbstractFormatter):
    """An PrimitiveEvent formatter."""

    def __init__(self) -> None:
        super().__init__()

    def format(self, data) -> PrimitiveEvent:
        """
        :param data: The data to format.
        :return: A new PrimitiveEvent instance containing the data.
        """
        return PrimitiveEvent(timestamp=EpochNSClock.generate_timestamp(),
                              data=data)



from abc import abstractmethod
from typing import Dict
from uuid import uuid4

from bobocep.rules.events.abstract_event import AbstractEvent


class BoboEvent(AbstractEvent):
    """A :code:`bobocep` event.

    :param timestamp: The event timestamp indicating when it was first
                      generated.
    :type timestamp: int

    :param data: The event data, defaults to an empty dict.
    :type data: Dict[str, str], optional

    :param event_id: The event ID, defaults to a randomly generated ID.
    :type event_id: str, optional
    """

    TIMESTAMP = "timestamp"
    DATA = "data"
    ID = "id"

    def __init__(self,
                 timestamp: int,
                 data: Dict[str, str] = None,
                 event_id: str = None) -> None:
        super().__init__()

        self.timestamp = timestamp
        self.data = {} if data is None else data
        self.id = '{}-{}'.format(uuid4(), timestamp) \
            if event_id is None else event_id

    @abstractmethod
    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """



from bobocep.decider.bobo_decider import BoboDecider
from bobocep.decider.bobo_decider_builder import BoboDeciderBuilder
from bobocep.decider.handlers.bobo_nfa_handler import BoboNFAHandler
from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.setup.distributed.incoming.dist_incoming_subscriber import \
    IDistIncomingSubscriber
from bobocep.setup.distributed.outgoing.dist_outgoing_subscriber import \
    IDistOutgoingSubscriber


class DistDecider(BoboDecider,
                  IDistIncomingSubscriber,
                  IDistOutgoingSubscriber):
    """A distributed data decider that enables synchronisation of Decider
    state across multiple instances of :code:`bobocep`.

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional
    """

    def __init__(self, max_queue_size: int = 0) -> None:
        super().__init__(max_queue_size=max_queue_size)

        self.is_synced = False

    def _loop(self) -> None:
        if self.is_synced:
            super()._loop()

    def on_sync(self) -> None:
        self.is_synced = True

    def on_receiver_event(self, event: BoboEvent) -> None:
        if self.is_synced:
            super().on_receiver_event(event)

    def on_accepted_producer_event(self, event: CompositeEvent) -> None:
        if self.is_synced:
            super().on_accepted_producer_event(event)

    def get_current_state(self) -> dict:
        """
        :return: The current state of the DistDecider instance, represented as
                 a dict.
        """

        return super().to_dict()

    def put_current_state(self, new_current_state: dict) -> None:
        """
        Updates the current state of the DistDecider instance with a new
        current state, represented as a dict.

        :param new_current_state: The new current state.
        :type new_current_state: dict
        """

        if self.is_synced:
            return

        with self._lock:
            for handler_dict in new_current_state[self.HANDLERS]:
                # find the corresponding nfa handler
                nfa_name = handler_dict[BoboNFAHandler.NFA_NAME]
                handler = self._nfa_handlers[nfa_name]

                # create buffer
                buffer = BoboDeciderBuilder.shared_versioned_match_buffer(
                    handler_dict[BoboNFAHandler.BUFFER])
                handler.buffer = buffer

                # overwrite existing runs
                handler.clear_runs(halt=False, notify=False)

                for run_dict in handler_dict[BoboNFAHandler.RUNS]:
                    run = BoboDeciderBuilder.run(run_dict, buffer, handler.nfa)
                    handler.add_run(run)

            self.is_synced = True

    def on_dist_run_transition(self,
                               nfa_name: str,
                               run_id: str,
                               state_name_from: str,
                               state_name_to: str,
                               event: BoboEvent):
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError

            handler.force_run_transition(
                run_id,
                state_name_from,
                state_name_to,
                event)

    def on_dist_run_clone(self,
                          nfa_name: str,
                          run_id: str,
                          next_state_name: str,
                          next_event: BoboEvent):
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError

            handler.force_run_clone(
                run_id,
                next_state_name,
                next_event)

    def on_dist_run_halt(self,
                         nfa_name: str,
                         run_id: str):
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError

            handler.force_run_halt(run_id)

    def on_dist_run_final(self,
                          nfa_name: str,
                          run_id: str,
                          history: BoboHistory) -> None:
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError

            handler.force_run_final(run_id=run_id, history=history)

    def on_sync_response(self, sync_id: str, body: str):
        """"""



from abc import ABC, abstractmethod


class AbstractReceiver(ABC):
    """An abstract data receiver."""

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def add_data(self, data) -> None:
        """
        :param data: Adds new data to the receiver.
        """



from abc import ABC

from bobocep.rules.actions.bobo_action import BoboAction


class MultiAction(BoboAction, ABC):
    """An abstract action that enables the execution of multiple actions
    within a single BoboAction instance."""

    def __init__(self) -> None:
        super().__init__()



import json
from queue import Queue
from time import time
from uuid import uuid4

import pika

import bobocep.setup.distributed.bobo_dist_constants as bdc
from bobocep.decider.dist_decider import DistDecider
from bobocep.decider.handlers.nfa_handler_subscriber import \
    INFAHandlerSubscriber
from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.setup.distributed.incoming.dist_incoming_subscriber import \
    IDistIncomingSubscriber
from bobocep.setup.distributed.outgoing.dist_outgoing_subscriber import \
    IDistOutgoingSubscriber
from bobocep.setup.task.bobo_task import BoboTask


class BoboDistOutgoing(BoboTask,
                       INFAHandlerSubscriber,
                       IDistIncomingSubscriber):
    """Handles outgoing synchronisation data for other :code:`bobocep`
    instances.

    :param decider: The decider to which synchronisation will occur.
    :type decider: DistDecider

    :param exchange_name: The exchange name to connect to on the external
                          message queue system.
    :type exchange_name: str

    :param user_id: The user ID to use on the external message queue system.
    :type user_id: str

    :param host_name: The host name of the external message queue system.
    :type host_name: str
    """

    def __init__(self,
                 decider: DistDecider,
                 exchange_name: str,
                 user_id: str,
                 host_name: str,
                 max_sync_attempts: int = 3) -> None:
        super().__init__()

        connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=host_name))
        channel = connection.channel()

        self.decider = decider
        self.exchange_name = exchange_name
        self.user_id = user_id
        self.sync_id = None
        self.host_name = host_name
        self.max_sync_attempts = max_sync_attempts

        self._queue_transition = Queue()
        self._queue_clone = Queue()
        self._queue_halt = Queue()
        self._queue_final = Queue()

        self._subs = []
        self._connection = connection
        self._channel = channel
        self._is_synced = False
        self._sync_response = None

    def _setup(self) -> None:
        self._sync()

    def _loop(self) -> None:
        if self._is_synced:
            self._send_events(self._queue_transition, bdc.TRANSITION)
            self._send_events(self._queue_clone, bdc.CLONE)
            self._send_events(self._queue_halt, bdc.HALT)
            self._send_events(self._queue_final, bdc.FINAL)

    def _cancel(self):
        self._is_synced = False
        self._connection.close()

    def on_sync_response(self, sync_id, body: str):
        if not self._cancelled:
            if self.sync_id == sync_id:
                self._sync_response = body

    def subscribe(self, subscriber: IDistOutgoingSubscriber) -> None:
        """
        :param subscriber: Subscribes to outgoing data.
        :type subscriber: IDistOutgoingSubscriber
        """

        if subscriber not in self._subs:
            self._subs.append(subscriber)

    def unsubscribe(self, unsubscriber: IDistOutgoingSubscriber) -> None:
        """
        :param unsubscriber: Unsubscribes to outgoing data.
        :type unsubscriber: IDistOutgoingSubscriber
        """

        if not self._cancelled:
            if unsubscriber in self._subs:
                self._subs.remove(unsubscriber)

    def _sync(self) -> None:
        sync_attempt = 0

        while not self._is_synced and sync_attempt < self.max_sync_attempts:
            sync_attempt += 1
            self._is_synced = self._sync_request()

        # assume nothing to sync (i.e. no other clients) if unsuccessful in
        # all attempts
        self._is_synced = True

        for subscriber in self._subs:
            subscriber.on_sync()

    def _sync_request(self, timeout: int = 3) -> bool:
        self._sync_response = None
        self.sync_id = str(uuid4())

        self._channel.basic_publish(
            exchange=self.exchange_name,
            routing_key=bdc.SYNC_REQ,
            properties=pika.BasicProperties(
                reply_to=bdc.SYNC_RES,
                message_id=self.user_id,
                correlation_id=self.sync_id,
            ),
            body=str())

        time_sent = time()

        # while no response and no timeout
        while self._sync_response is None and (time() - time_sent) < timeout:
            self._connection.process_data_events()

        # if timeout (i.e. no response), assume there are no other clients
        if self._sync_response is None:
            return False

        self.decider.put_current_state(json.loads(self._sync_response))

        return True

    def _send_events(self, queue: Queue, routing_key: str):
        if not queue.empty():
            data = queue.get_nowait()

            if data is not None:
                data_json = json.dumps(data)
                self._channel.basic_publish(
                    exchange=self.exchange_name,
                    properties=pika.spec.BasicProperties(
                        message_id=self.user_id),
                    routing_key=routing_key,
                    body=data_json
                )

    def on_handler_transition(self,
                              nfa_name: str,
                              run_id: str,
                              state_name_from: str,
                              state_name_to: str,
                              event: BoboEvent):
        if not self._cancelled:
            self._queue_transition.put_nowait({
                bdc.NFA_NAME: nfa_name,
                bdc.RUN_ID: run_id,
                bdc.STATE_FROM: state_name_from,
                bdc.STATE_TO: state_name_to,
                bdc.EVENT: event.to_dict()
            })

    def on_handler_clone(self,
                         nfa_name: str,
                         run_id: str,
                         state_name: str,
                         event: BoboEvent):
        if not self._cancelled:
            self._queue_clone.put_nowait({
                bdc.NFA_NAME: nfa_name,
                bdc.RUN_ID: run_id,
                bdc.STATE_TO: state_name,
                bdc.EVENT: event.to_dict()
            })

    def on_handler_halt(self,
                        nfa_name: str,
                        run_id: str):
        if not self._cancelled:
            self._queue_halt.put_nowait({
                bdc.NFA_NAME: nfa_name,
                bdc.RUN_ID: run_id
            })

    def on_handler_final(self,
                         nfa_name: str,
                         run_id: str,
                         event: CompositeEvent):
        if not self._cancelled:
            self._queue_final.put_nowait({
                bdc.NFA_NAME: nfa_name,
                bdc.RUN_ID: run_id,
                bdc.EVENT: event.to_dict()
            })

    def on_dist_run_transition(self,
                               nfa_name: str,
                               run_id: str,
                               state_name_from: str,
                               state_name_to: str,
                               event: BoboEvent):
        """"""

    def on_dist_run_clone(self,
                          nfa_name: str,
                          run_id: str,
                          next_state_name: str,
                          next_event: BoboEvent):
        """"""

    def on_dist_run_halt(self,
                         nfa_name: str,
                         run_id: str):
        """"""

    def on_dist_run_final(self,
                          nfa_name: str,
                          run_id: str,
                          event: BoboEvent) -> None:
        """"""



from abc import abstractmethod

from bobocep.producer.producer_subscriber import IProducerSubscriber
from bobocep.rules.actions.abstract_action import AbstractAction
from bobocep.rules.events.composite_event import CompositeEvent


class BoboAction(AbstractAction,
                 IProducerSubscriber):
    """A :code:`bobocep` action."""

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def perform_action(self, event: CompositeEvent) -> bool:
        """
        Performs an action using the event provided.

        :param event: A composite event to use as part of the action process.
        :type event: CompositeEvent

        :return: True if the action was successful, False otherwise.
        """

    def on_accepted_producer_event(self, event: CompositeEvent) -> None:
        self.perform_action(event)

    def on_rejected_producer_event(self, event: CompositeEvent) -> None:
        """"""



from abc import ABC, abstractmethod

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory


class IDistIncomingSubscriber(ABC):

    @abstractmethod
    def on_dist_run_transition(self,
                               nfa_name: str,
                               run_id: str,
                               state_name_from: str,
                               state_name_to: str,
                               event: BoboEvent) -> None:
        """
        When a transition has occurred on a run on another :code:`bobocep`
        instance.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The ID of the run that transitioned.
        :type run_id: str

        :param state_name_from: The original state.
        :type state_name_from: str

        :param state_name_to: The new state.
        :type state_name_to: str

        :param event: The event that caused the transition.
        :type event: BoboEvent
        """

    @abstractmethod
    def on_dist_run_clone(self,
                          nfa_name: str,
                          run_id: str,
                          next_state_name: str,
                          next_event: BoboEvent) -> None:
        """
        When a clone has occurred on a run on another :code:`bobocep` instance.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The ID of the cloned run.
        :type run_id: str

        :param next_state_name: The state of the cloned run.
        :type next_state_name: str

        :param next_event: The event of the cloned run.
        :type next_event: BoboEvent
        """

    @abstractmethod
    def on_dist_run_halt(self,
                         nfa_name: str,
                         run_id: str) -> None:
        """
        When a run has halted on a run on another :code:`bobocep` instance.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The ID of the halted run.
        :type run_id: str
        """

    @abstractmethod
    def on_dist_run_final(self,
                          nfa_name: str,
                          run_id: str,
                          history: BoboHistory) -> None:
        """
        When a run has reached its final state on on another :code:`bobocep`
        instance.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The ID of the accepted run.
        :type run_id: str

        :param history: The history of the accepted run.
        :type history: BoboHistory
        """

    @abstractmethod
    def on_sync_response(self,
                         sync_id: str,
                         body: str) -> None:
        """
        The response from another :code:`bobocep` instance when attempting to
        synchronise decider state across instances.

        :param sync_id: The ID of the synchronise request.
        :type sync_id: str

        :param body: The body of the response containing synchronise data.
        :type body: str
        """



from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.predicates.windows.sliding.window_sliding import \
    WindowSliding


class WindowSlidingLast(WindowSliding):
    """
    A predicate that evaluates using a sliding time window,
    where the window ranges from the last event in history to the next event.

    :param interval_sec: The sliding window time interval, in seconds.
    :type interval_sec: float
    """

    def __init__(self, interval_sec: float) -> None:
        super().__init__(interval_ns=int(interval_sec * 1e9))

    def get_previous_event(self, history: BoboHistory) -> BoboEvent:
        return history.last



from random import randint

from bobocep.receiver.generators.bobo_null_data import BoboNullData


class BoboNullDataRandInt(BoboNullData):
    """
    Null data class that generates a random int within a given range.

    :param imin: The minimum possible int value.
    :type imin: int

    :param imax: The maximum possible int value.
    :type imax: int

    :raises RuntimeError: Value of imin must be lower than imax.
    """

    def __init__(self, imin: int, imax: int) -> None:
        super().__init__()

        if imin >= imax:
            raise RuntimeError("Value of imin {} must be lower than imax {}."
                               .format(imin, imax))

        self._imin = imin
        self._imax = imax

    def get_null_data(self):
        return randint(self._imin, self._imax)



from bobocep.rules.actions.bobo_action import BoboAction
from bobocep.rules.events.composite_event import CompositeEvent


class NoAction(BoboAction):
    """
    An action that does nothing and always returns the specified boolean value.

    :param bool_return: The boolean value to always return when performing
                        the action, defaults to True.
    :type bool_return: bool, optional
    """

    def __init__(self, bool_return: bool = True) -> None:
        super().__init__()

        self._bool_return = bool_return

    def perform_action(self, event: CompositeEvent) -> bool:
        return self._bool_return



from abc import ABC

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.events.primitive_event import PrimitiveEvent
from bobocep.rules.nfas.bobo_nfa import BoboNFA
from bobocep.rules.nfas.patterns.bobo_pattern import BoboPattern
from bobocep.rules.states.bobo_state import BoboState
from bobocep.rules.transitions.bobo_transition import BoboTransition


class BoboRuleBuilder(ABC):
    """Builds rule types from serialized representations of those types."""

    @staticmethod
    def event(d: dict) -> BoboEvent:
        """
        Creates either a PrimitiveEvent or CompositeEvent instance from a dict
        representation of one of them.

        :param d: A dict representation of either a PrimitiveEvent or a
                  CompositeEvent instance.
        :type d: dict

        :return: Either a PrimitiveEvent or CompositeEvent instance, both of
                 which are returned by their common superclass BoboEvent.
        """

        return BoboRuleBuilder.composite(d) if CompositeEvent.NAME in d \
            else BoboRuleBuilder.primitive(d)

    @staticmethod
    def primitive(d: dict) -> 'PrimitiveEvent':
        """Creates a PrimitiveEvent instance from serialized representation of
           one.

        :param d: A dict representation of a PrimitiveEvent instance.
        :type d: dict

        :raises RuntimeError: Timestamp not found in dict.
        :raises RuntimeError: Data not found in dict.

        :return: A PrimitiveEvent instance.
        """

        timestamp = d.get(BoboEvent.TIMESTAMP)

        if timestamp is None:
            raise RuntimeError("Timestamp not found in dict.")

        data = d.get(BoboEvent.DATA)

        if data is None:
            raise RuntimeError("Data not found in dict.")

        event_id = d.get(BoboEvent.ID)

        if event_id is None:
            raise RuntimeError("Event ID not found in dict.")

        return PrimitiveEvent(timestamp=timestamp,
                              data=data,
                              event_id=event_id)

    @staticmethod
    def composite(d: dict) -> 'CompositeEvent':
        """Creates a CompositeEvent instance from serialized representation of
           one.

        :param d: A dict representation of a CompositeEvent instance.
        :type d: dict

        :raises RuntimeError: Timestamp not found in dict.
        :raises RuntimeError: Name not found in dict.
        :raises RuntimeError: History not found in dict.

        :return: A CompositeEvent instance.
        """

        timestamp = d.get(BoboEvent.TIMESTAMP)

        if timestamp is None:
            raise RuntimeError("Timestamp not found in dict.")

        name = d.get(CompositeEvent.NAME)

        if name is None:
            raise RuntimeError("Name not found in dict.")

        history_dict = d.get(CompositeEvent.HISTORY)

        if history_dict is None:
            raise RuntimeError("History not found in dict.")

        history = BoboRuleBuilder.history(history_dict)

        data = d.get(BoboEvent.DATA)

        event_id = d.get(BoboEvent.ID)

        if event_id is None:
            raise RuntimeError("Event ID not found in dict.")

        return CompositeEvent(timestamp=timestamp,
                              name=name,
                              history=history,
                              data=data,
                              event_id=event_id)

    @staticmethod
    def history(d: dict) -> 'BoboHistory':
        """Creates a BoboHistory instance from serialized representation of
           one.

        :param d: A dict representation of a BoboHistory instance.
        :type d: dict

        :return: A BoboHistory instance.
        """

        event_history_dict = {}

        for key in d.keys():
            key_events = d[key]

            event_history_dict[key] = []

            for event_dict in key_events:
                event_history_dict[key].append(
                    BoboRuleBuilder.event(event_dict))

        return BoboHistory(events=event_history_dict)

    @staticmethod
    def nfa(name_nfa: str, pattern: BoboPattern) -> BoboNFA:
        """Generates a named NFA from a pattern.

        :param name_nfa: The name for the automaton instance.
        :type name_nfa: str

        :param pattern: The pattern to use for the automaton definition.
        :type pattern: BoboPattern

        :raises RuntimeError: Pattern does not contain any layers.
        :raises RuntimeError: The start and the final state must not
                              contain loops, be negated, be optional, or be
                              nondeterministic.
        :raises RuntimeError: State names must be unique.

        :return: A new BoboNFA instance.
        """

        if len(pattern.layers) == 0:
            raise RuntimeError("Pattern does not contain any layers.")

        nfa_states = {}
        nfa_transitions = {}
        nfa_start_state_name = None
        nfa_final_state_name = None
        nfa_preconditions = pattern.preconditions
        nfa_haltconditions = pattern.haltconditions

        all_state_names = []
        last_state_names = None
        last_states = None
        last_layer = None
        len_layers = len(pattern.layers)

        for i, layer in enumerate(pattern.layers):

            # start / final states must be "normal"
            if (i == 0 or i == len_layers - 1) and (layer.loop or layer.negated
                                                    or layer.optional or
                                                    len(layer.predicates) > 1):
                raise RuntimeError("The start and final states must "
                                   "not contain loops, be negated, be "
                                   "optional, or be nondeterministic")

            for j in range(layer.times):
                state_names = []
                states = []

                for k, predicate in enumerate(layer.predicates):
                    state_name = "{}-{}-{}".format(layer.label, k + 1, j + 1)

                    state_names.append(state_name)
                    states.append(BoboState(state_name,
                                            layer.label,
                                            predicate,
                                            negated=layer.negated,
                                            optional=layer.optional))

                # first state is the start state
                if i == 0 and j == 0:
                    nfa_start_state_name = state_names[0]

                # last state is the final state
                if i == len_layers - 1 and j == layer.times - 1:
                    nfa_final_state_name = state_names[-1]

                # duplicate state name found
                for state_name in state_names:
                    if state_name in all_state_names:
                        raise RuntimeError("State names must be unique. "
                                           "Multiple uses found for name {}."
                                           .format(state_name))

                # add state name to list of current state names
                all_state_names.extend(state_names)

                # add states
                for state in states:
                    nfa_states[state.name] = state

                # add transitions for previous state(s)
                if last_state_names is not None and last_states is not None \
                        and last_layer is not None:

                    # include last layer's state (would only be one) if looping
                    if last_layer.loop:
                        state_names.extend(last_state_names)

                    # create transition and apply it to the previous state(s)
                    transition = BoboTransition(state_names=state_names,
                                                strict=layer.strict)
                    for state in last_states:
                        nfa_transitions[state.name] = transition

                last_state_names = state_names
                last_states = states
                last_layer = layer

        # add empty transition for final state
        nfa_transitions[nfa_final_state_name] = BoboTransition(
            state_names=[], strict=False)

        return BoboNFA(name=name_nfa,
                       states=nfa_states,
                       transitions=nfa_transitions,
                       start_state_name=nfa_start_state_name,
                       final_state_name=nfa_final_state_name,
                       preconditions=nfa_preconditions,
                       haltconditions=nfa_haltconditions)



from copy import copy
from threading import RLock
from typing import List

from bobocep.decider.buffers.shared_versioned_match_buffer import \
    SharedVersionedMatchBuffer
from bobocep.decider.runs.abstract_run import AbstractRun
from bobocep.decider.runs.run_subscriber import IRunSubscriber
from bobocep.decider.versions.run_version import RunVersion
from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.nfas.bobo_nfa import BoboNFA
from bobocep.rules.states.bobo_state import BoboState


class BoboRun(AbstractRun):
    """A :code:`bobocep` antomaton run.

    :param buffer: The buffer in which run data will be stored..
    :type buffer: SharedVersionedMatchBuffer

    :param nfa: The automaton with which the run is associated.
    :type nfa: BoboNFA

    :param event: The run event.
    :type event: BoboEvent

    :param start_time: The time when the run was created. It is used for
                       generating the run ID. Defaults to the event
                       instance's timestamp.
    :type start_time: int, optional

    :param start_state: The start state of the run, defaults to the NFA's start
                        state.
    :type start_state: BoboState, optional

    :param current_state: The current state of the run, defaults to the NFA's
                          start state.
    :type current_state: BoboState, optional

    :param parent_run: The run which is this run's parent. It is used when a
                       run has been created as a consequence of non-determinism
                       (i.e. run cloning). Defaults to None.
    :type parent_run: BoboRun, optional

    :param run_id: The ID of the run. Defaults to a run ID consisting of the
                   NFA name and start time.
    :type run_id: str, optional

    :param version: The run version. If a parent run is provided, the version
                    is generated relative to the parent version.
                    Defaults to the run ID.
    :type version: RunVersion, optional

    :param put_event: Puts the run event into the buffer on instantiation,
                      defaults to True.
    :type put_event: bool, optional

    :param last_process_cloned: The last time the run processed an event,
                                a clone occurred, defaults to False.
    :type last_process_cloned: bool, optional

    :param halted: The run is halted, defaults to False.
    :type halted: bool, optional
    """

    NFA_NAME = "nfa_name"
    EVENT = "event"
    START_TIME = "start_time"
    START_STATE_NAME = "start_state_name"
    CURRENT_STATE_NAME = "current_state_name"
    RUN_ID = "run_id"
    VERSION = "version"
    HALTED = "halted"
    LAST_PROCESS_CLONED = "last_process_cloned"

    def __init__(self,
                 buffer: SharedVersionedMatchBuffer,
                 nfa: BoboNFA,
                 event: BoboEvent,
                 start_time: int = None,
                 start_state: BoboState = None,
                 current_state: BoboState = None,
                 parent_run: 'BoboRun' = None,
                 run_id: str = None,
                 version: RunVersion = None,
                 put_event: bool = True,
                 last_process_cloned: bool = False,
                 halted: bool = False) -> None:

        super().__init__()

        self.buffer = buffer
        self.nfa = nfa
        self.event = event
        self.start_time = start_time if start_time is not None \
            else event.timestamp
        self.start_state = start_state if start_state is not None \
            else self.nfa.start_state
        self.current_state = current_state if current_state is not None \
            else self.start_state
        self.id = BoboRun._generate_id(
            nfa_name=self.nfa.name,
            start_event_id=self.event.id) \
            if run_id is None else run_id
        self.version = version
        self._last_process_cloned = last_process_cloned
        self._halted = halted
        self._final = False
        self._subs = []
        self._lock = RLock()

        if parent_run is None:
            # create new version and add new event under this version
            if self.version is None:
                self.version = RunVersion()
                self.version.add_level(self.id)

            if put_event:
                self.buffer.put_event(
                    nfa_name=self.nfa.name,
                    run_id=self.id,
                    version=self.version.get_version_as_str(),
                    state_label=self.current_state.label,
                    event=self.event)

        else:
            # create version from existing, and
            # put and link last event to new version
            if self.version is None:
                self.version = RunVersion(
                    parent_version=parent_run.version)
                self.version.add_level(self.id)

            if put_event:
                self.buffer.put_event(
                    nfa_name=self.nfa.name,
                    run_id=parent_run.id,
                    version=parent_run.version.get_version_as_str(),
                    state_label=self.current_state.label,
                    event=self.event,
                    new_run_id=self.id,
                    new_version=self.version.get_version_as_str())

        # immediately final if start state is final state
        if self.nfa.start_is_final:
            self._set_final(history=None, notify=False)

    @staticmethod
    def _generate_id(nfa_name: str, start_event_id: str) -> str:
        """
        Generates a run ID.

        :param nfa_name: The run NFA name.
        :type nfa_name: str

        :param start_event_id: The ID of the first event in the run,
        :type start_event_id: str

        :return: A run ID.
        """

        return "{}-{}".format(nfa_name, start_event_id)

    def process(self, event: BoboEvent, recent: List[CompositeEvent]) -> None:
        """
        Process an event.

        :param event: The event to process.
        :type event: BoboEvent

        :param recent: Recently accepted complex events of the corresponding
                        automaton.
        :type recent: List[CompositeEvent]

        :raises RuntimeError: Run has already halted.
        """
        with self._lock:
            if self._halted:
                raise RuntimeError("Run {} has already halted."
                                   .format(self.id))

            # get the history of the current run
            history = self.buffer.get_all_events(
                self.nfa.name,
                self.id,
                self.version)

            if self._any_preconditions_failed(event, history, recent) or \
                    self._any_haltconditions_passed(event, history, recent):
                self.set_halt()
            else:
                self._handle_state(self.current_state, event, history, recent)

    def last_process_cloned(self) -> bool:
        """
        :return: True if the last time the run processed an event, a clone
                 occurred, False otherwise.
        """

        return self._last_process_cloned

    def is_halted(self) -> bool:
        """
        :return: True if the run has halted, False otherwise.
        """

        with self._lock:
            return self._halted

    def is_final(self) -> bool:
        """
        :return: True if run has reached its final state, False otherwise.
        """

        with self._lock:
            return self._final

    def set_halt(self, notify: bool = True) -> None:
        """
        Halt the run.

        :param notify: Whether to notify run subscribers of the halting,
                       defaults to True.
        :type notify: bool, optional
        """

        with self._lock:
            if not self._halted:
                self._halted = True

                if notify:
                    self._notify_halt()

    def subscribe(self, subscriber: IRunSubscriber) -> None:
        """
        :param subscriber: Subscribes to the run.
        :type subscriber: IRunSubscriber
        """

        with self._lock:
            if subscriber not in self._subs:
                self._subs.append(subscriber)

    def unsubscribe(self, unsubscriber: IRunSubscriber) -> None:
        """
        :param unsubscriber: Unsubscribes from the run.
        :type unsubscriber: IRunSubscriber

        :raises RuntimeError: Run has already halted.
        """

        with self._lock:
            if unsubscriber in self._subs:
                self._subs.remove(unsubscriber)

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        with self._lock:
            return {
                self.NFA_NAME: self.nfa.name,
                self.EVENT: self.event.to_dict(),
                self.START_TIME: self.start_time,
                self.START_STATE_NAME: self.start_state.name,
                self.CURRENT_STATE_NAME: self.current_state.name,
                self.RUN_ID: self.id,
                self.VERSION: copy(self.version._levels),
                self.HALTED: self._halted,
                self.LAST_PROCESS_CLONED: self._last_process_cloned
            }

    def _set_final(self,
                   history: BoboHistory = None,
                   notify: bool = True) -> None:
        if notify:
            self._notify_final(
                history if history is not None else
                self.buffer.get_all_events(
                    self.nfa.name,
                    self.id,
                    self.version))

        # do not notify halt if final
        self.set_halt(notify=False)

        self._final = True

    def _handle_state(self,
                      state: BoboState,
                      event: BoboEvent,
                      history: BoboHistory,
                      recent: List[CompositeEvent]) -> None:
        transition = self.nfa.transitions.get(state.name)

        if transition is None:
            raise RuntimeError("No transition found for state {}."
                               .format(state.name))

        for trans_state_name in transition.state_names:
            # get transition state from NFA
            trans_state = self.nfa.states[trans_state_name]

            # state successfully fulfilled
            if trans_state.process(event, history, recent):
                # negated i.e. should NOT have occurred, so halt
                if trans_state.is_negated:
                    self.set_halt()
                    break

                if not transition.is_deterministic:
                    # not a self loop: clone run
                    if trans_state_name != state.name:
                        self._notify_clone(trans_state, event)
                        self._last_process_cloned = True
                    else:
                        # increment current run
                        history = self._proceed(
                            event=event,
                            original_state=state,
                            trans_state=trans_state,
                            increment=self._last_process_cloned)
                        self._last_process_cloned = False
                else:
                    # deterministic: proceed as normal
                    history = self._proceed(event, state, trans_state)
            else:
                # if optional, or if state is negated: move to the next state
                if transition.is_deterministic and \
                        (trans_state.is_optional or trans_state.is_negated):
                    self._handle_state(trans_state, event, history, recent)

                # halt if requires strict contiguity
                elif transition.is_strict:
                    self.set_halt()
                    break

    def _proceed(self,
                 event: BoboEvent,
                 original_state: BoboState,
                 trans_state: BoboState,
                 increment: bool = False,
                 notify: bool = True) -> BoboHistory:

        if original_state.name not in self.nfa.states.keys():
            raise RuntimeError(
                "Original state {} not in NFA {}.".format(
                    original_state.name, self.nfa.name))

        if trans_state.name not in self.nfa.states.keys():
            raise RuntimeError(
                "Transition state {} not in NFA {}.".format(
                    original_state.name, self.nfa.name))

        if increment:
            new_increment = BoboRun._generate_id(
                nfa_name=self.nfa.name,
                start_event_id=event.id)
            new_version = self.version.list_to_version_str(
                self.version.get_version_as_list()[:-1] + [new_increment])
        else:
            new_increment = None
            new_version = None

        # add new event to buffer
        self.buffer.put_event(
            nfa_name=self.nfa.name,
            run_id=self.id,
            version=self.version.get_version_as_str(),
            state_label=trans_state.label,
            event=event,
            new_version=new_version)

        # (maybe) apply increment
        if new_increment is not None:
            self.version.increment_level(new_increment)

        # get run history
        new_history = self.buffer.get_all_events(
            self.nfa.name,
            self.id,
            self.version)

        # halt if final, else transition
        if self.nfa.final_state.name == trans_state.name:
            self._set_final(new_history, notify=notify)

        elif notify:
            self._notify_transition(
                original_state.name,
                trans_state.name,
                event)

        # update run
        self.current_state = trans_state
        self.event = event

        return new_history

    def _any_preconditions_failed(self,
                                  event: BoboEvent,
                                  history: BoboHistory,
                                  recent: List[CompositeEvent]) -> bool:
        """If any preconditions are False, return True."""

        return any(not p.evaluate(event, history, recent)
                   for p in self.nfa.preconditions)

    def _any_haltconditions_passed(self,
                                   event: BoboEvent,
                                   history: BoboHistory,
                                   recent: List[CompositeEvent]) -> bool:
        """If any haltconditions are True, return True."""

        return any(p.evaluate(event, history, recent)
                   for p in self.nfa.haltconditions)

    def _notify_transition(self,
                           state_name_from: str,
                           state_name_to: str,
                           event: BoboEvent) -> None:
        for subscriber in self._subs:
            subscriber.on_run_transition(
                run_id=self.id,
                state_name_from=state_name_from,
                state_name_to=state_name_to,
                event=event,
                notify=True)

    def _notify_clone(self,
                      state: BoboState,
                      next_event: BoboEvent) -> None:
        for subscriber in self._subs:
            subscriber.on_run_clone(
                state_name=state.name,
                event=next_event,
                parent_run_id=self.id,
                force_parent=False,
                notify=True)

    def _notify_final(self,
                      history: BoboHistory) -> None:
        for subscriber in self._subs:
            subscriber.on_run_final(
                run_id=self.id,
                history=history,
                halt=False,
                notify=True)

    def _notify_halt(self) -> None:
        for subscriber in self._subs:
            subscriber.on_run_halt(
                run_id=self.id,
                notify=True)



from abc import ABC

from bobocep.decider.buffers.match_event import MatchEvent
from bobocep.decider.buffers.shared_versioned_match_buffer import \
    SharedVersionedMatchBuffer
from bobocep.decider.runs.bobo_run import BoboRun
from bobocep.decider.versions.run_version import RunVersion
from bobocep.rules.bobo_rule_builder import BoboRuleBuilder
from bobocep.rules.nfas.bobo_nfa import BoboNFA


class BoboDeciderBuilder(ABC):
    """A builder for classes related to the :code:`bobocep` decision making
    subsystems.
    """

    @staticmethod
    def match_event(d: dict) -> 'MatchEvent':
        """
        :param d: A dict representation of a MatchEvent instance.
        :type d: dict

        :return: A new MatchEvent instance.
        """

        return MatchEvent(nfa_name=d[MatchEvent.NFA_NAME],
                          label=d[MatchEvent.LABEL],
                          event=BoboRuleBuilder.event(d[MatchEvent.EVENT]),
                          next_ids=d[MatchEvent.NEXT_IDS],
                          previous_ids=d[MatchEvent.PREVIOUS_IDS])

    @staticmethod
    def shared_versioned_match_buffer(d: dict) -> SharedVersionedMatchBuffer:
        """
        :param d: A dict representation of a SharedVersionedMatchBuffer
                  instance.
        :type d: dict

        :return: A new SharedVersionedMatchBuffer instance.
        """

        buffer = SharedVersionedMatchBuffer()

        for eve_dict in d[SharedVersionedMatchBuffer.EVENTS]:
            nfa_name = eve_dict[SharedVersionedMatchBuffer.NFA_NAME]
            match_event = BoboDeciderBuilder.match_event(
                eve_dict[SharedVersionedMatchBuffer.MATCH_EVENT])

            nfa_labels = SharedVersionedMatchBuffer._get_or_create_subdict(
                buffer._eve, nfa_name)
            nfa_events = SharedVersionedMatchBuffer._get_or_create_subdict(
                nfa_labels, match_event.label)
            nfa_events[match_event.event.id] = match_event

        for ver_dict in d[SharedVersionedMatchBuffer.LAST]:
            nfa_name = ver_dict[SharedVersionedMatchBuffer.NFA_NAME]
            label = ver_dict[SharedVersionedMatchBuffer.LABEL]
            run_id = ver_dict[SharedVersionedMatchBuffer.RUN_ID]
            version = ver_dict[SharedVersionedMatchBuffer.VERSION]
            event_id = ver_dict[SharedVersionedMatchBuffer.EVENT_ID]

            match_event = buffer._eve[nfa_name][label][event_id]

            nfa_runs = SharedVersionedMatchBuffer._get_or_create_subdict(
                buffer._ver, nfa_name)
            run_versions = SharedVersionedMatchBuffer._get_or_create_subdict(
                nfa_runs, run_id)
            run_versions[version] = match_event

        return buffer

    @staticmethod
    def run(d: dict,
            buffer: SharedVersionedMatchBuffer,
            nfa: BoboNFA) -> 'BoboRun':
        """
        :param d: A dict representation of a BoboRun instance.
        :type d: dict

        :param buffer: A buffer to use with the new BoboRun instance.
        :type buffer: SharedVersionedMatchBuffer

        :param nfa: An automaton to use with the new BoboRun instance.
        :type nfa: BoboNFA

        :return: A new BoboRun instance.
        """

        event = BoboRuleBuilder.event(d[BoboRun.EVENT])
        start_time = d[BoboRun.START_TIME]
        start_state = nfa.states[d[BoboRun.START_STATE_NAME]]
        current_state = nfa.states[d[BoboRun.CURRENT_STATE_NAME]]
        run_id = d[BoboRun.RUN_ID]
        version = RunVersion.list_to_version(d[BoboRun.VERSION])
        last_proceed_had_clone = d[BoboRun.LAST_PROCESS_CLONED]
        halted = d[BoboRun.HALTED]

        return BoboRun(buffer=buffer,
                       nfa=nfa,
                       event=event,
                       start_time=start_time,
                       start_state=start_state,
                       current_state=current_state,
                       run_id=run_id,
                       version=version,
                       put_event=False,
                       last_process_cloned=last_proceed_had_clone,
                       halted=halted)



from typing import List

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.predicates.bobo_predicate import BoboPredicate
from bobocep.rules.states.abstract_state import AbstractState


class BoboState(AbstractState):
    """A :code:`bobocep` state for use in automata.

    :param name: The state name.
    :type name: str

    :param label: The state label for grouping multiple states.
    :type label: str

    :param predicate: The state predicate that, if True, would cause a
                      transition to this state.
    :type predicate: BoboPredicate

    :param negated: Whether the state's predicate should *not* happen
                    (i.e. predicate passes evaluation if False),
                    defaults to False.
    :type negated: bool, optional

    :param optional: Whether the state is optional
                     (i.e. will not halt the corresponding automaton
                     if predicate returns False)
                     defaults to False.
    :type optional: bool, optional
    """

    def __init__(self,
                 name: str,
                 label: str,
                 predicate: BoboPredicate,
                 negated: bool = False,
                 optional: bool = False) -> None:
        super().__init__()

        self.name = name
        self.label = label
        self.predicate = predicate
        self.is_negated = negated
        self.is_optional = optional

    def process(self,
                event: BoboEvent,
                history: BoboHistory,
                recent: List[CompositeEvent]) -> bool:
        """Evaluates the state predicate.

        :param event: The event used in the evaluation.
        :type event: BoboEvent

        :param history: The events previously accepted by earlier the state's
                        corresponding automaton.
        :type history: BoboHistory

        :param recent: Recently accepted complex events of the corresponding
                        automaton.
        :type recent: List[CompositeEvent]

        :return: True if the state's predicate evaluates to True,
                 False otherwise.
        :rtype: bool
        """

        return self.predicate.evaluate(event, history, recent)
